package de.latzig.tool.lcdcontroller;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class LCDControllerVT52Font {

    private static final String PREFIX_LNE = ";#.db";

    private static final Pattern PATTERN_DATA = Pattern
            .compile("^0x([0-9A-F]{2}),0x([0-9A-F]{2}),0x([0-9A-F]{2}),0x([0-9A-F]{2}),0x([0-9A-F]{2}),0x([0-9A-F]{2}),0x([0-9A-F]{2}),0x([0-9A-F]{2}).*0x([0-9A-F]{2}).*$");

    private static final String FORMAT_DATA = "%s 0x%02X,0x%02X,0x%02X,0x%02X,0x%02X,0x%02X,0x%02X,0x%02X   ; 0x%02X";

    /**
     * Entry point of CLI.
     * 
     * @param args
     *            The args of CLI.
     */
    public static void main(String[] args) {
        try {
            LCDControllerVT52Font prog = new LCDControllerVT52Font();
            if (args.length < 1) {
                System.out.println(String.format("Usage: %s [File]", prog
                        .getClass().getSimpleName()));
                System.exit(-1);
                return;
            }
            prog.perform(args);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(-1);
        }
    }

    /**
     * Performs program.
     * 
     * @param args
     *            The args of CLI.
     * @throws Exception
     *             While performing program.
     */
    private void perform(String[] args) throws Exception {
        // get file
        String filePath = args[0];
        File file = new File(filePath);
        if (!file.isFile()) {
            // invalid binary file
            throw new IllegalArgumentException(String.format(
                    "Invalid file: %s", file.getAbsolutePath()));
        }
        System.out.println(String.format("File: %s", file.getAbsolutePath()));

        // read font data from text file
        System.out.println("Read font data...");
        int[] font = new int[256 * 8];
        BufferedReader br = new BufferedReader(new FileReader(file));
        try {
            String line = null;
            int i = 1;
            int c = 0x20;
            while ((line = br.readLine()) != null) {
                line = line.trim();
                if (line.startsWith(PREFIX_LNE)) {
                    line = line.substring(PREFIX_LNE.length()).trim();
                    Matcher matcher = PATTERN_DATA.matcher(line);
                    if (!matcher.matches()) {
                        // invalid binary file
                        throw new IllegalArgumentException(String.format(
                                "Invalid data in line: %d", Integer.valueOf(i)));
                    }
                    int t = Integer.valueOf(matcher.group(9), 16);
                    if (t != c) {
                        throw new IllegalArgumentException(
                                String.format("Invalid index in line: %d",
                                        Integer.valueOf(i)));
                    }
                    for (int p = 0; p < 8; p++) {
                        int v = Integer.valueOf(matcher.group(p + 1), 16);
                        font[8 * (c - 0x20) + p] = v;
                    }
                    c++;
                }
                i++;
            }
        } finally {
            br.close();
        }

        // invert characters on 0x80-0xFF
        System.out.println("Invert upper half of font...");
        for (int c = 0x20; c < 0xA0; c++) {
            for (int p = 0; p < 8; p++) {
                int v = font[8 * (c - 0x020) + p];
                v = v ^ 0xFF;
                font[8 * (c - 0x20 + 0x80) + p] = v;
            }
        }

        // write new text file with calculated font data
        System.out.println("Write new font data...");
        BufferedWriter bw = new BufferedWriter(new FileWriter(file));
        try {
            // write original font data
            bw.append("FONT:\r\n");
            for (int c = 0x20; c < 0xA0; c++) {
                int v0 = font[8 * (c - 0x20) + 0];
                int v1 = font[8 * (c - 0x20) + 1];
                int v2 = font[8 * (c - 0x20) + 2];
                int v3 = font[8 * (c - 0x20) + 3];
                int v4 = font[8 * (c - 0x20) + 4];
                int v5 = font[8 * (c - 0x20) + 5];
                int v6 = font[8 * (c - 0x20) + 6];
                int v7 = font[8 * (c - 0x20) + 7];

                String line = String.format(FORMAT_DATA, PREFIX_LNE,
                        Integer.valueOf(v0), Integer.valueOf(v1),
                        Integer.valueOf(v2), Integer.valueOf(v3),
                        Integer.valueOf(v4), Integer.valueOf(v5),
                        Integer.valueOf(v6), Integer.valueOf(v7),
                        Integer.valueOf(c));
                bw.append(line);
                bw.append("\r\n");
            }

            // write calculated font data
            for (int p = 0; p < 8; p++) {
                bw.append("\r\n.db ");
                for (int c = 0x20; c < 0xA0; c++) {
                    int v = font[8 * (c - 0x20) + p];
                    bw.append(String.format("0x%02X,", Integer.valueOf(v)));
                }
                for (int c = 0x20; c < 0xA0; c++) {
                    int v = font[8 * (c - 0x20 + 0x80) + p];
                    bw.append(String.format("0x%02X", Integer.valueOf(v)));
                    if (c != 0x9F) {
                        bw.append(",");
                    }
                }
            }
        } finally {
            bw.close();
        }
    }
}
