#ifndef __IDE_INC
#define __IDE_INC

.include "util.inc"
.include "mcp23s17.inc"

; #############################################################################
; ###                                                                       ###
; ###                     IDE Interface (via MCP23S17)                      ###  
; ###                                                                       ###
; #############################################################################

; *****************************************************************************
; *** Definitions                                                           ***
; *****************************************************************************

.equ IDE_SECTOR_SIZE = 512				; Size of a sector

.equ IDE_REG_DATA = 0x10				; CS0+ADDR %000: Data Register
.equ IDE_REG_ERROR = 0x11				; CS0+ADDR %001: Error Register
.equ IDE_BIT_ERROR_AMNF = (1<<0)		; Error Register Bit0 (AMNF): Indicates that an address mark was not found.
.equ IDE_BIT_ERROR_TK0NF = (1<<1)		; Error Register Bit1 (TK0NF): When this bit is set the drive was not able to find track 0 of the device.
.equ IDE_BIT_ERROR_ABRT = (1<<2)		; Error Register Bit2 (ABRT): Command was aborted due to drive status error or invalid command.
.equ IDE_BIT_ERROR_MCR = (1<<3)			; Error Register Bit3 (MCR): Media change was requested.
.equ IDE_BIT_ERROR_IDNF = (1<<4)		; Error Register Bit4 (IDNF): Sector ID was not found (wrong sector number).
.equ IDE_BIT_ERROR_MC = (1<<5)			; Error Regitser Bit5 (MC): Indicates that the media has been changed.
.equ IDE_BIT_ERROR_UNC = (1<<6)			; Error Register Bit6 (UNC): Uncorrectable data error. Set when the sector data can't be recreated (even with ECC).
.equ IDE_BIT_ERROR_BAD = (1<<7)			; Error Register Bit7 (BAD): Bad block. This bit is set when the requested sectors ID contained a bad block mark.
.equ IDE_REG_SECTOR_COUNT = 0x12		; CS0+ADDR %010: Sector Counter Register
.equ IDE_REG_SECTOR = 0x13				; CS0+ADDR %011: Start Sector Register
.equ IDE_REG_CYLINDER_LO = 0x14			; CS0+ADDR %100: Lo-byte of the Cylinder Number
.equ IDE_REG_CYLINDER_HI = 0x15			; CS0+ADDR %101: Hi-byte of the Cylinder Number
.equ IDE_REG_SDH = 0x16					; CS0+ADDR %110: Head and Device Select Register
.equ IDE_BIT_SDH_HEAD0 = (1<<0)			; SDH Register Bit0 (HEAD0): Head (0-15)
.equ IDE_BIT_SDH_HEAD1 = (1<<1)			; SDH Register Bit1 (HEAD1): Head (0-15)
.equ IDE_BIT_SDH_HEAD2 = (1<<2)			; SDH Register Bit2 (HEAD2): Head (0-15)
.equ IDE_BIT_SDH_HEAD3 = (1<<3)			; SDH Register Bit3 (HEAD3): Head (0-15)
.equ IDE_BIT_SDH_DRIVE = (1<<4)			; SDH Register Bit4 (DRIVE): Drive Select (0=Master/1=Slave)
.equ IDE_BIT_SDH_SIZE0 = (1<<5)			; SDH Register Bit5 (SIZE0): Sector Size (01=512Byte OOTB, 11=512Byte LBA Mode)
.equ IDE_BIT_SDH_SIZE1 = (1<<6)			; SDH Register Bit6 (SIZE1): Sector Size (01=512Byte OOTB, 11=512Byte LBA Mode) 
.equ IDE_BIT_SDH_CRC = (1<<7)			; SDH Register Bit7 (CRC): Historical: Extension Bit. When zero, CRC data is appended to the sectors data fields. When set to one, no CRC data is appended. Default is 1.
.equ IDE_REG_CMD_STATUS = 0x17			; CS0+ADDR %111: Command/Status Register
.equ IDE_BIT_STATUS_ERR = (1<<0)		; Status Register Bit0 (ERR): Error Bit. If this bit is set then an error has occurred while executing the latest command. The error status itself is to be found in the error register.
.equ IDE_BIT_STATUS_IND = (1<<1)		; Status Register Bit1 (IND): Index Bit. This bit is active once per disk revolution. May be used to determine rotational speed.
.equ IDE_BIT_STATUS_ECC = (1<<2)		; Status Register Bit2 (ECC): Corrected data flag. Set when there was a correctable data error and the data has been corrected.
.equ IDE_BIT_STATUS_DRQ = (1<<3)		; Status Register Bit3 (DRQ): Data request. This bit indicates that the drive is ready for a data transfer.
.equ IDE_BIT_STATUS_SKC = (1<<4)		; Status Register Bit4 (SKC): Drive seek complete. This bit is set when the actuator of the drives head is on track. This bit also is updated similarly to drive ready.
.equ IDE_BIT_STATUS_WFT = (1<<5)		; Status Register Bit5 (WFT): Drive write fault. Similar to drive ready, this bit is not updated after an error.
.equ IDE_BIT_STATUS_RDY = (1<<6)		; Status Register Bit6 (RDY): Drive ready. This bit is set when the drive is up to speed and ready to accept a command. When there is an error, this bit is not updated until the next read of the status register, so it can be used to determine the cause of the error.
.equ IDE_BIT_STATUS_BSY = (1<<7)		; Status Register Bit7 (BSY): Busy flag. When this flag is set, the task file registers must not be accessed due to internal operations.
.equ IDE_REG_ALTERNATE = 0x26			; CS1+ADDR %110: Alternate Status (w/o setting IRQ)
.equ IDE_REG_DRIVE_ADDR = 0x27			; CS1+ADDR %111: Drive Address

; ### COMMANDS
; (SC=Sector Count, SN=Sector Number, C=Cylinder, H=Head, D=Data Register) 
.equ IDE_CMD_READ = 0x20				; CMD [$20]: Read Sector with Retry using SC,SN,C,D,H.
.equ IDE_CMD_READ_NO_RETRY = 0x21		; CMD [$21]: Read Sector without Retry using SC,SN,C,D,H.
.equ IDE_CMD_WRITE = 0x30				; CMD [$30]: Write Sector with Retry using SC,SN,C,D,H.
.equ IDE_CMD_WRITE_NO_RETRY = 0x31		; CMD [$31]: Write Sector without Retry using SC,SN,C,D,H.
.equ IDE_CMD_IDENTIFY = 0xEC			; CMD [$EC]: Identify Drive using D.

; ### IDENTIFY BLOCK
.equ IDE_ID_CONF = 0x000				; ID [$000]: Configuration/ID word
.equ IDE_BIT_CONF_NON_MAGNETIC = (1<<15); Configuration Bit 15: Non-magnetic drive
.equ IDE_BIT_CONF_FMT_SPD_GAP = (1<<14) ; Configuration Bit 14: Format speed tolerance gap required
.equ IDE_BIT_CONF_TRACK_OFFSET = (1<<13); Configuration Bit 13: Track offset option available
.equ IDE_BIT_CONF_DATA_STROBE = (1<<12)	; Configuration Bit 12: Data strobe offset option available
.equ IDE_BIT_CONF_ROT_FREQ = (1<<11)	; Configuration Bit 11: Rotational frequency tolerance > 0.5%
.equ IDE_BIT_CONF_DAT_TRN_10MB = (1<<10); Configuration Bit 10: Data transfer rate > 10 MB/s
.equ IDE_BIT_CONF_DAT_TRN_5MB = (1<<9)	; Configuration Bit 9: Data transfer rate > 5 MB/s, <= 10 MB/s
.equ IDE_BIT_CONF_DAT_TRN = (1<<8)		; Configuration Bit 8: Data transfer rate <= 5 MB/s
.equ IDE_BIT_CONF_REMOVABLE = (1<<7)	; Configuration Bit 7: Removable disk
.equ IDE_BIT_CONF_NON_REMOVABLE = (1<<6); Configuration Bit 6: Non-removable disk
.equ IDE_BIT_CONF_MOTOR_OFF = (1<<5)	; Configuration Bit 5: Spindle motor can be switched off
.equ IDE_BIT_CONF_HEAD_SWITCH = (1<<4)	; Configuration Bit 4: Head switching time > 15 us
.equ IDE_BIT_CONF_NOT_MFM_ENC = (1<<3)	; Configuration Bit 3: Not MFM encoded
.equ IDE_BIT_CONF_SOFT_SECTORED = (1<<2); Configuration Bit 2: Soft sectored
.equ IDE_BIT_CONF_HARD_SECTORED = (1<<1); Configuration Bit 1: Hard sectored
.equ IDE_ID_FIXED_CYLINDERS = 0x002		; ID [$002]: Number of fixed cylinders
.equ IDE_ID_REMOVABLE_CYLINDERS = 0x004	; ID [$004]: Number of removable cylinders
.equ IDE_ID_HEADS = 0x006				; ID [$006]: Number of heads
.equ IDE_ID_BYTES_PHYS_TRACKS = 0x008	; ID [$008]: Number of unformatted bytes per physical track
.equ IDE_ID_BYTES_SECTOR = 0x00A		; ID [$00A]: Number of unformatted bytes per sector
.equ IDE_ID_PHYS_SECTORS_TRACKS = 0x00C	; ID [$00C]: Number of phyical sectors per Tracks
.equ IDE_ID_BYTES_SECTOR_GAPS = 0x00E	; ID [$00E]: Number of of bytes in the inter-sector gaps
.equ IDE_ID_BYTES_SYNC_FIELDS = 0X010	; ID [$010]: Number of bytes in the sync fields 
.equ IDE_ID_SERIAL_NUMBER = 0x014		; ID [$014]: Serial number (20 chars)
.equ IDE_ID_CONTROLLER_TYPE = 0x028		; ID [$028]: Controller type 
.equ IDE_VAL_CTRL_TYPE_UNKNOWN = 0		; Controller Type 0=Not specified
.equ IDE_VAL_CTRL_TYPE_SINGLE_PORTED = 1; Controller Type 1=Single ported, single sector buffer
.equ IDE_VAL_CTRL_TYPE_DUAL_PORTED = 2	; Controller Type 2=Dual ported, multiple sector buffer
.equ IDE_VAL_CTRL_TYPE_DUAL_PORTED2 = 3	; Controller Type 3=Dual ported, multiple sector buffer, with look-ahead read capabilities)
.equ IDE_ID_CONTROLLER_BUF_SIZE = 0x02A	; ID [$02A]: Controller buffer size (in sectors)
.equ IDE_ID_ECC_BYTES = 0x02C			; ID [$02C]: Number of of ECC bytes on long commands
.equ IDE_ID_CONTROLLER_FIRMWARE = 0x02E ; ID [$02E]: Controller firmware revision (8 chars)
.equ IDE_ID_MODEL_NUMBER = 0x036		; ID [$036]: Model number (40 chars)
.equ IDE_ID_SECTORS_INTERRUPT = 0x05E	; ID [$05E]: Number of of sectors/interrupt (0 = no support)
.equ IDE_ID_DOUBLE_WORD_TRANSFER = 0x060; ID [$060]: Double word transfer flag (1 = capable)
.equ IDE_ID_WRITE_PROTECTED = 0x062		; ID [$062]: Write protected

#ifndef IDE_SPI
; ### PINS (Control Pins by AVR)
#ifndef IDE_PINS
#define IDE_PINS
.equ IDE_PIN_A0 = (1<<0)				; PC0: A0 (Adresss Bit 0)
.equ IDE_PIN_A1 = (1<<1)				; PC1: A1 (Address Bit 1)
.equ IDE_PIN_A2 = (1<<2)				; PC2: A2 (Address Bit 2)
.equ IDE_PIN_CS0 = (1<<3)				; PC3: CS0 (ChipSelect #0, low-active)
.equ IDE_PIN_CS1 = (1<<4)				; PC3: CS1 (ChipSelect #1, low-active)
.equ IDE_PIN_RD = (1<<0)				; PB0: RD (Read, low-active)
.equ IDE_PIN_WR = (1<<1)				; PB1: WR (Write, low-active)

.equ IDE_PORTDAT_ADDR = PORTC			; A0-A2, CS0 & CS1
.equ IDE_PORTDIR_ADDR = DDRC
.equ IDE_PORTDAT_RW = PORTB				; RD & WR
.equ IDE_PORTDIR_RW = DDRB
#endif
#else 
; ### PINS (Control Pins by MCP23S17)
#ifndef IDE_SPI_PINS
#define IDE_SPI_PINS
.equ IDE_SPI_PIN_A0 = (1<<0)			; Pin A0: A0 (Adresss Bit 0)
.equ IDE_SPI_PIN_A1 = (1<<1)			; Pin A1: A1 (Address Bit 1)
.equ IDE_SPI_PIN_A2 = (1<<2)			; Pin A2: A2 (Address Bit 2)
.equ IDE_SPI_PIN_CS0 = (1<<3)			; Pin A3: CS0 (ChipSelect #0, low-active)
.equ IDE_SPI_PIN_CS1 = (1<<4)			; Pin A4: CS1 (ChipSelect #1, low-active)
.equ IDE_SPI_PIN_RD = (1<<5)			; Pin A5: RD (Read, low-active)
.equ IDE_SPI_PIN_WR = (1<<6)			; Pin A6: WR (Write, low-active)
.equ IDE_SPI_PIN_LED = (1<<7)			; Pin A7: LED
#endif
#endif

.ifndef IDE_GPIO_ADDR
.equ IDE_GPIO_ADDR = MCP23S17_ADDR0		; MCP23S17 Address for data transfer
.endif
#ifdef IDE_SPI
.ifndef IDE_GPIO_ADDR2
.equ IDE_GPIO_ADDR2 = MCP23S17_ADDR1	; MCP23S17 Address for control pins (only used for SPI IDE)
.endif
#endif

.dseg              
; *****************************************************************************
; *** SRAM                                                                  ***
; *****************************************************************************
IDE_BUFFER:		.byte IDE_SECTOR_SIZE	; Buffer for one sector

.cseg
; *****************************************************************************
; *** DATA                                                                  ***
; *****************************************************************************

; *****************************************************************************
; *** Public functions                                                      ***
; *****************************************************************************

; ### Inits IDE Interface
#ifndef IDE_FN_INIT
#define IDE_FN_INIT
ide_init:
		push a
		push b
		push c
		push pdat
		push zl
		push zh

		lds a,UTIL_INIT1					; Check, if initialization has been already done
		sbrc a,INIT1_IDE		
		rjmp ide_init_done

		mcall util_init						; Init depending modules
		mcall mcp23s17_init						
		
#ifndef IDE_SPI
		; ### PINS (Control Pins by AVR)
		in a,IDE_PORTDIR_ADDR				; Prepare address pins
		sbr a,IDE_PIN_A0|IDE_PIN_A1|IDE_PIN_A2|IDE_PIN_CS0|IDE_PIN_CS1
		out IDE_PORTDIR_ADDR,a
		in pdat,IDE_PORTDAT_ADDR
		cbr pdat,IDE_PIN_A0|IDE_PIN_A1|IDE_PIN_A2
		sbr pdat,IDE_PIN_CS0|IDE_PIN_CS1
		out IDE_PORTDAT_ADDR,pdat			

		in a,IDE_PORTDIR_RW					; Prepare r/w pins
		sbr a,IDE_PIN_RD|IDE_PIN_WR
		out IDE_PORTDIR_RW,a
		in pdat,IDE_PORTDAT_RW
		sbr a,IDE_PIN_RD|IDE_PIN_WR
		out IDE_PORTDAT_RW,pdat			
#else 
		; ### PINS (Control Pins by MCP23S17)
		ldi b,IDE_GPIO_ADDR2|MCP23S17_REG_IOCON		
		mcall mcp23s17_read
		sbr a,MCP23S17_IOCON_HAEN
		mcall mcp23s17_write

		ldi b,IDE_GPIO_ADDR2				; Prepare control pins (for SPI IDE only)
		mcall mcp23s17_get_iodir_a
		cbr a,IDE_SPI_PIN_A0|IDE_SPI_PIN_A1|IDE_SPI_PIN_A2|IDE_SPI_PIN_CS0|IDE_SPI_PIN_CS1|IDE_SPI_PIN_RD|IDE_SPI_PIN_WR								
.if IDE_SPI_PIN_LED<=0x80
		cbr a,IDE_SPI_PIN_LED
.endif
		mcall mcp23s17_set_iodir_a
		mcall mcp23s17_get_gpio_a	
		cbr a,IDE_SPI_PIN_A0|IDE_SPI_PIN_A1|IDE_SPI_PIN_A2
		sbr a,IDE_SPI_PIN_RD|IDE_SPI_PIN_WR|IDE_SPI_PIN_CS0|IDE_SPI_PIN_CS1
.if IDE_SPI_PIN_LED<=0x80
		cbr a,IDE_SPI_PIN_LED				
.endif
		mcall mcp23s17_set_gpio_a	
#endif
		; ### PINS (Address Pins by MCP23S17)
		ldi b,IDE_GPIO_ADDR|MCP23S17_REG_IOCON		
		mcall mcp23s17_read
		sbr a,MCP23S17_IOCON_HAEN
		mcall mcp23s17_write

		ldi b,IDE_GPIO_ADDR					; Prepare data pins
		ser a								
		mcall mcp23s17_set_iodir_a
		mcall mcp23s17_set_iodir_b
		mcall mcp23s17_set_pullup_a
		mcall mcp23s17_set_pullup_b
		clr a
		mcall mcp23s17_set_gpio_a	
		mcall mcp23s17_set_gpio_b	
		
ide_init_clear:								; Clears buffer for (Sector) Data
		rcall ide_clear

ide_init_done:
		lds a,UTIL_INIT1					; Mark as initialized
		sbr a,(1<<INIT1_IDE)
		sts UTIL_INIT1,a

		pop zh
		pop zl
		pop pdat
		pop c
		pop b
		pop a
		ret
#endif

; ### Clears Buffer (for Sector)
#ifndef IDE_FN_CLEAR
#define IDE_FN_CLEAR
ide_clear:
		push a
		push c
		push zl
		push zh

		ldi zl,LOW(IDE_BUFFER)
		ldi zh,HIGH(IDE_BUFFER)
		clr c
ide_clear_loop:
		clr a
		st z+,a
		st z+,a
		dec c
		brne ide_clear_loop	

		pop zh
		pop zl
		pop c
		pop a
		ret
#endif

; ### Perform "Identify" command (Fills buffer with identification data from HDD)
; ><pdat: The drive selection (0=Master, 1=Slave) and the error code as output (zero on no error)
#ifndef IDE_FN_IDENTIFY
#define IDE_FN_IDENTIFY
ide_identify:
		push a
		push b
		push c
		push d
		
		rcall _ide_led_on					; Enable LED

		rcall _ide_wait_until_ready			; Waits until ready
		clr a								; Setup LBA
		clr b
		clr c
		clr d
		rcall _ide_set_lba
		ldi a,IDE_CMD_IDENTIFY				; Perform command "Identify" (0xEC)
		ldi b,IDE_REG_CMD_STATUS		
		rcall _ide_set_reg

		rcall _ide_wait_for_data_request	; Waits for data request
		rcall _ide_check_error				; Check for error
		tst pdat
		brne ide_identify_done
		rcall _ide_read_data				; Reads data into buffer

ide_identify_done:
		rcall _ide_led_off					; Disable LED

ide_identify_exit:
		pop d
		pop c
		pop b
		pop a
		ret
#endif

; ### Perform "Read Sector" command
; >a: LBA Bits 0-7
; >b: LBA Bits 8-15
; >c: LBA Bits 16-23
; >d: LBA Bits 24-27 
; ><pdat: The drive selection (0=Master, 1=Slave) and the error code as output (zero on no error)
#ifndef IDE_FN_READ
#define IDE_FN_READ
ide_read:
		push a
		push b
		push c
		push d

		rcall _ide_led_on					; Enable LED

		rcall _ide_wait_until_ready			; Waits until ready
		rcall _ide_set_lba					; Setup LBA
		ldi a,1								; Set Counter for one Sector
		ldi b,IDE_REG_SECTOR_COUNT
		rcall _ide_set_reg
		ldi a,IDE_CMD_READ					; Perform command "Read Sector" (0x20)
		ldi b,IDE_REG_CMD_STATUS		
		rcall _ide_set_reg

		rcall _ide_wait_for_data_request	; Waits for data request
		rcall _ide_check_error				; Check for error
		tst pdat
		brne ide_read_done
		rcall _ide_read_data				; Reads data into buffer

ide_read_done:
		rcall _ide_led_off					; Disable LED

ide_read_exit:
		pop d
		pop c
		pop b
		pop a
		ret
#endif

; ### Perform "Write Sector" command
; >a: LBA Bits 0-7
; >b: LBA Bits 8-15
; >c: LBA Bits 16-23
; >d: LBA Bits 24-27 
; ><pdat: The drive selection (0=Master, 1=Slave) and the error code as output (zero on no error)
#ifndef IDE_FN_WRITE
#define IDE_FN_WRITE
ide_write:
		push a
		push b
		push c
		push d

		rcall _ide_led_on					; Enable LED

		rcall _ide_wait_until_ready			; Waits until ready
		rcall _ide_set_lba					; Setup LBA
		ldi a,1								; Set Counter for one Sector
		ldi b,IDE_REG_SECTOR_COUNT
		rcall _ide_set_reg
		ldi a,IDE_CMD_WRITE					; Perform command "Write Sector" (0x30)
		ldi b,IDE_REG_CMD_STATUS		
		rcall _ide_set_reg

		rcall _ide_wait_for_data_request	; Waits for data request
		rcall _ide_write_data				; Writes data from buffer
		rcall _ide_wait_until_ready			; Waits until ready
		rcall _ide_check_error				; Check for error

ide_write_done:
		rcall _ide_led_off					; Disable LED

ide_write_exit:
		pop d
		pop c
		pop b
		pop a
		ret
#endif

; ### Reads Byte from Buffer
; <a: The Byte
; >o: Offset inside Buffer
ide_buf_get:
		push zl
		push zh

		clr a								; Check Range of Offset
		cpi oh,HIGH(IDE_SECTOR_SIZE)
		brsh ide_buf_get_exit

		ldi zl,LOW(IDE_BUFFER)				; Read Byte
		ldi zh,HIGH(IDE_BUFFER)
		add zl,ol
		adc zh,oh
		ld a,z

ide_buf_get_exit:
		pop zh
		pop zl
		ret

; ### Writes Byte to Buffer
; >a: The Byte
; >o: Offset inside Buffer
ide_buf_set:
		push zl
		push zh

		cpi oh,HIGH(IDE_SECTOR_SIZE)		; Check Range of Offset
		brsh ide_buf_set_exit

		ldi zl,LOW(IDE_BUFFER)				; Read Byte
		ldi zh,HIGH(IDE_BUFFER)
		add zl,ol
		adc zh,oh
		st z,a

ide_buf_set_exit:
		pop zh
		pop zl
		ret

; *****************************************************************************
; *** Private functions                                                     ***
; *****************************************************************************

; ### Reads data (for sector) into buffer
#ifndef IDE_FN__READ_DATA
#define IDE_FN__READ_DATA
_ide_read_data:
		push a
		push b
		push c
		push zl
		push zh

		ldi a,IDE_REG_DATA					; Sets address of the register
		rcall _ide_set_addr

		ldi b,IDE_GPIO_ADDR					; Switch data pins to input
		ser a								
		mcall mcp23s17_set_iodir_a
		mcall mcp23s17_set_iodir_b
		
		ldi zl,LOW(IDE_BUFFER)				; Set pointer for buffer
		ldi zh,HIGH(IDE_BUFFER)				
		clr c								; Resets counter for 256 words (512 byte for buffer)
_ide_read_data_loop:
		rcall _ide_activate_read			; Activate Read-Mode
		mcall mcp23s17_get_gpio_b			; Reads 16bit-data
		st z+,a
		mcall mcp23s17_get_gpio_a
		st z+,a
		rcall _ide_deactivate				; Deactive Read-Write-Modes
		dec c								
		brne _ide_read_data_loop

		pop zh
		pop zl
		pop c
		pop b
		pop a
		ret
#endif

; ### Writes data (for sector) from buffer
#ifndef IDE_FN__WRITE_DATA
#define IDE_FN__WRITE_DATA
_ide_write_data:
		push a
		push b
		push c
		push zl
		push zh

		ldi a,IDE_REG_DATA					; Sets address of the register
		rcall _ide_set_addr

		ldi b,IDE_GPIO_ADDR					; Switch data pins to output
		clr a								
		mcall mcp23s17_set_iodir_a
		mcall mcp23s17_set_iodir_b
		
		ldi zl,LOW(IDE_BUFFER)				; Set pointer for buffer
		ldi zh,HIGH(IDE_BUFFER)				
		clr c								; Resets counter for 256 words (512 byte for buffer)
_ide_write_data_loop:
		ld a,z+								; Writes 16bit-data
		mcall mcp23s17_set_gpio_b
		ld a,z+
		mcall mcp23s17_set_gpio_a
		rcall _ide_activate_write			; Activate Write-Mode
		rcall _ide_deactivate				; Deactive Read-Write-Modes
		dec c								
		brne _ide_write_data_loop

		pop zh
		pop zl
		pop c
		pop b
		pop a
		ret
#endif

; ### Set LBA Bits 0-27
; >a: LBA Bits 0-7
; >b: LBA Bits 8-15
; >c: LBA Bits 16-23
; >d: LBA Bits 24-27 
; pdat>  The drive selection (0=Master, 1=Slave)
#ifndef IDE_FN__SET_LBA
#define IDE_FN__SET_LBA
_ide_set_lba:
		push temp
		push a
		push b
		push c
		push d
		
		mov temp,b							; Set LBA Bits 0-7
		ldi b,IDE_REG_SECTOR
		rcall _ide_set_reg
		mov a,temp							; Set LBA Bits 8-15
		ldi b,IDE_REG_CYLINDER_LO			
		rcall _ide_set_reg
		mov a,c								; Set LBA Bits 16-23
		ldi b,IDE_REG_CYLINDER_HI
		rcall _ide_set_reg					
		mov a,d								; Set LBA Bits 24-27 (mark SDH for LBA mode)
		andi a,IDE_BIT_SDH_HEAD0|IDE_BIT_SDH_HEAD1|IDE_BIT_SDH_HEAD2|IDE_BIT_SDH_HEAD3
		sbr a,IDE_BIT_SDH_SIZE0|IDE_BIT_SDH_SIZE1|IDE_BIT_SDH_CRC
		cpse pdat,zero						; Check drive selection 
		sbr a,IDE_BIT_SDH_DRIVE				; Is not zero, set to DRIVE bit to Slave
		ldi b,IDE_REG_SDH
		rcall _ide_set_reg					
	
		pop d
		pop c
		pop b
		pop a
		pop temp
		ret
#endif

; ### Check for error code
; <pdat: The error code, or zero on no error
#ifndef IDE_FN__CHECK_ERROR
#define IDE_FN__CHECK_ERROR
_ide_check_error:
		push a

		clr pdat							; Reset error code
		rcall _ide_get_status				; Check for ERR=1
		andi a,IDE_BIT_STATUS_ERR
		tst a
		breq _ide_check_error_exit
		rcall _ide_get_error				; Yes, there is an error
		mov pdat,a							; Store error
_ide_check_error_exit:
		pop a
		ret
#endif

; ### Wait until ready for data request
#ifndef IDE_FN__WAIT_FOR_DATA_REQUEST
#define IDE_FN__WAIT_FOR_DATA_REQUEST
_ide_wait_for_data_request:
		push a
			
_ide_wait_for_data_request_loop:			; Wait until BSY=0 and RDY=1
		rcall _ide_get_status				
		andi a,IDE_BIT_STATUS_DRQ|IDE_BIT_STATUS_BSY
		cpi a,IDE_BIT_STATUS_DRQ
		brne _ide_wait_for_data_request_loop				
		
		pop a	
		ret
#endif

; ### Wait until ready
#ifndef IDE_FN__WAIT_UNTIL_READY
#define IDE_FN__WAIT_UNTIL_READY
_ide_wait_until_ready:
		push a
			
_ide_wait_until_ready_loop:					; Wait until BSY=0 and RDY=1
		rcall _ide_get_status				
		andi a,IDE_BIT_STATUS_RDY|IDE_BIT_STATUS_BSY
		cpi a,IDE_BIT_STATUS_RDY
		brne _ide_wait_until_ready_loop				
		
		pop a	
		ret
#endif

; ### Reads status register
#ifndef IDE_FN__GET_STATUS_REGISTER
#define IDE_FN__GET_STATUS_REGISTER
_ide_get_status:
		push b

		ldi b,IDE_REG_CMD_STATUS
		rcall _ide_get_reg

		pop b
		ret
#endif

; ### Reads error register
#ifndef IDE_FN__GET_ERROR_REGISTER
#define IDE_FN__GET_ERROR_REGISTER
_ide_get_error:
		push b

		ldi b,IDE_REG_ERROR
		rcall _ide_get_reg

		pop b
		ret
#endif

; ### Reads register
; <a: The value read from register
; >b: The address selection (see IDE_REG_xxx).
#ifndef IDE_FN__GET_REG
#define IDE_FN__GET_REG
_ide_get_reg:
		push b
		
		mov a,b								; Sets address of the register
		rcall _ide_set_addr

		ldi b,IDE_GPIO_ADDR					; Switch data pins to input
		ser a								
		mcall mcp23s17_set_iodir_a
		mcall mcp23s17_set_iodir_b
		
		rcall _ide_activate_read			; Activate Read-Mode
		mcall mcp23s17_get_gpio_a			; Reads 8bit-data (D0-D7)
		rcall _ide_deactivate				; Deactive Read-Write-Modes

		pop b
		ret
#endif

; ### Writes register
; >a: The value to write to register
; >b: The address selection (see IDE_REG_xxx).
#ifndef IDE_FN__SET_REG
#define IDE_FN__SET_REG
_ide_set_reg:
		push temp
		push a
		push b

		mov temp,a
		mov a,b								; Sets address of the register
		rcall _ide_set_addr

		ldi b,IDE_GPIO_ADDR					; Switch data pins to output
		clr a								
		mcall mcp23s17_set_iodir_a
		mcall mcp23s17_set_iodir_b
		
		mov a,temp							; Write 8bit-data (D0-D7)
		mcall mcp23s17_set_gpio_a			
		rcall _ide_activate_write			; Activate Write-Mode
		rcall _ide_deactivate				; Deactive Read-Write-Modes

		pop b
		pop a
		pop temp
		ret
#endif

; ### Sets address (A0-A2, CS0 & CS1)
; >a: The address selection (see IDE_REG_xxx).
#ifndef IDE_FN__SET_ADDR
#define IDE_FN__SET_ADDR
_ide_set_addr:
#ifndef IDE_SPI
; ### PINS (Control Pins by AVR)
		push pdat

		in pdat,IDE_PORTDAT_ADDR			; Modify current address selection
		cbr pdat,IDE_PIN_A0|IDE_PIN_A1|IDE_PIN_A2
		sbr pdat,IDE_PIN_CS0|IDE_PIN_CS1
		sbrc a,0							; Check A0 (Bit 0)
		sbr pdat,IDE_PIN_A0
		sbrc a,1							; Check A1 (Bit 1)
		sbr pdat,IDE_PIN_A1
		sbrc a,2							; Check A2 (Bit 2)
		sbr pdat,IDE_PIN_A2
		sbrc a,4							; Check CS0 (Bit 4)
		cbr pdat,IDE_PIN_CS0
		sbrc a,5							; Check CS1 (Bit 5)
		cbr pdat,IDE_PIN_CS1
		out IDE_PORTDAT_ADDR,pdat

		pop  pdat
		ret
#else
; ### PINS (Control Pins by MCP23S17)
		push temp
		push a
		push b
		
		mov temp,a
		ldi b,IDE_GPIO_ADDR2				; Modify current address selection
		mcall mcp23s17_get_gpio_a			
		cbr a,IDE_SPI_PIN_A0|IDE_SPI_PIN_A1|IDE_SPI_PIN_A2
		sbr a,IDE_SPI_PIN_CS0|IDE_SPI_PIN_CS1
		sbrc temp,0							; Check A0 (Bit 0)
		sbr a,IDE_SPI_PIN_A0
		sbrc temp,1							; Check A1 (Bit 1)
		sbr a,IDE_SPI_PIN_A1
		sbrc temp,2							; Check A2 (Bit 2)
		sbr a,IDE_SPI_PIN_A2
		sbrc temp,4							; Check CS0 (Bit 4)
		cbr a,IDE_SPI_PIN_CS0
		sbrc temp,5							; Check CS1 (Bit 5)
		cbr a,IDE_SPI_PIN_CS1
		mcall mcp23s17_set_gpio_a	

		pop b
		pop a
		pop temp
		ret
#endif
#endif

; ### Activate Read Mode
#ifndef IDE_FN__ACTIVATE_READ
#define IDE_FN__ACTIVATE_READ
_ide_activate_read:
#ifndef IDE_SPI
; ### PINS (Control Pins by AVR)
		push pdat

		in pdat,IDE_PORTDAT_RW
		sbr pdat,IDE_PIN_WR
		cbr pdat,IDE_PIN_RD
		out IDE_PORTDAT_RW,pdat			

		pop pdat
		ret
#else
; ### PINS (Control Pins by MCP23S17)
		push a
		push b
		
		ldi b,IDE_GPIO_ADDR2				; Modify current address selection
		mcall mcp23s17_get_gpio_a			
		sbr a,IDE_SPI_PIN_WR
		cbr a,IDE_SPI_PIN_RD
		mcall mcp23s17_set_gpio_a	

		pop b
		pop a
		ret
#endif
#endif

; ### Activate Write Mode
#ifndef IDE_FN__ACTIVATE_WRITE
#define IDE_FN__ACTIVATE_WRITE
_ide_activate_write:
#ifndef IDE_SPI
; ### PINS (Control Pins by AVR)
		push pdat

		in pdat,IDE_PORTDAT_RW
		sbr pdat,IDE_PIN_RD
		cbr pdat,IDE_PIN_WR
		out IDE_PORTDAT_RW,pdat			

		pop pdat
		ret
#else
; ### PINS (Control Pins by MCP23S17)
		push a
		push b
		
		ldi b,IDE_GPIO_ADDR2				; Modify current address selection
		mcall mcp23s17_get_gpio_a			
		sbr a,IDE_SPI_PIN_RD
		cbr a,IDE_SPI_PIN_WR
		mcall mcp23s17_set_gpio_a	

		pop b
		pop a
		ret
#endif
#endif

; ### Deactivate Read-Write Modes
#ifndef IDE_FN__DEACTIVATE
#define IDE_FN__DEACTIVATE
_ide_deactivate:
#ifndef IDE_SPI
; ### PINS (Control Pins by AVR)
		push pdat

		in pdat,IDE_PORTDAT_RW
		sbr pdat,IDE_PIN_RD|IDE_PIN_WR
		out IDE_PORTDAT_RW,pdat			

		pop pdat
		ret
#else
; ### PINS (Control Pins by MCP23S17)
		push a
		push b
		
		ldi b,IDE_GPIO_ADDR2				; Modify current address selection
		mcall mcp23s17_get_gpio_a			
		sbr a,IDE_SPI_PIN_RD|IDE_SPI_PIN_WR
		mcall mcp23s17_set_gpio_a	

		pop b
		pop a
		ret
#endif
#endif

; ### Active LED
#ifndef IDE_FN__LED_ON
#define IDE_FN__LED_ON
_ide_led_on:
#ifndef IDE_SPI
; ### PINS (Control Pins by AVR)
		; (nothing to do by default)
		ret		
#else
; ### PINS (Control Pins by MCP23S17)
.if IDE_SPI_PIN_LED<=0x80
		push a
		push b
		
		ldi b,IDE_GPIO_ADDR2				; Modify current address selection
		mcall mcp23s17_get_gpio_a			
		sbr a,IDE_SPI_PIN_LED				
		mcall mcp23s17_set_gpio_a	

		pop b
		pop a
.endif
		ret
#endif
#endif

; ### Deactive LED
#ifndef IDE_FN__LED_OFF
#define IDE_FN__LED_OFF
_ide_led_off:
#ifndef IDE_SPI
; ### PINS (Control Pins by AVR)
		; (nothing to do by default)
		ret		
#else
; ### PINS (Control Pins by MCP23S17)
.if IDE_SPI_PIN_LED<=0x80
		push a
		push b
		
		ldi b,IDE_GPIO_ADDR2				; Modify current address selection
		mcall mcp23s17_get_gpio_a			
		cbr a,IDE_SPI_PIN_LED				
		mcall mcp23s17_set_gpio_a	

		pop b
		pop a
.endif
		ret
#endif
#endif

#endif ; __IDE_INC
