#ifndef __MATH_INC
#define __MATH_INC

.include "util.inc"

; #############################################################################
; ###                                                                       ###
; ###                     MATH (Math Library for 32 Bit)                    ###  
; ###                                                                       ###
; #############################################################################

; *****************************************************************************
; *** Definitions                                                           ***
; *****************************************************************************

; === CONSTANTS ===============================================================

.ifndef MATH_VAR_NUM						; Number of Variables in Buffer
.equ MATH_VAR_NUM = 16
.endif
.equ MATH_VAR_MAX = 64						; Maximal Number of Variables for Buffer (4*64 = 256 Bytes)
.if MATH_VAR_NUM > MATH_VAR_MAX
.error "Out of Range vor Numer of Variables (max. 64)!"
.endif

.equ MATH_VAR_BUF_SIZE = 4*MATH_VAR_NUM		; Size of Buffer for Variables

.dseg              
; *****************************************************************************
; *** SRAM                                                                  ***
; *****************************************************************************

.if MATH_VAR_NUM>0
MATH_VARS:	.byte MATH_VAR_BUF_SIZE			; Buffer for Variables
.endif

.cseg
; *****************************************************************************
; *** DATA                                                                  ***
; *****************************************************************************

; *****************************************************************************
; *** Public functions                                                      ***
; *****************************************************************************

; ### Inits Math Library
#ifndef MATH_FN_INIT
#define MATH_FN_INIT
math_init:
		push a

		sbrc init0,INIT0_MATH				; Check, if initialization has been already done
		rjmp math_init_done
		
		mcall util_init						; Init depending modules
					
		rcall math_clear_a					; Reset accu & operand
		rcall math_clear_o

.if MATH_VAR_NUM>0
		rcall math_var_clear				; Reset Buffer for Variables
.endif

math_init_done:
		mov a,init0							; Mark as initialized
		sbr a,(1<<INIT0_MATH)
		mov init0,a

		pop a
		ret
#endif

; ### Clears Buffer for Variables
#ifndef MATH_FN_VAR_CLEAR
#define MATH_FN_VAR_CLEAR
.if MATH_VAR_NUM>0
math_var_clear:
		push a
		push b
		push zl
		push zh

		ldi zl,LOW(MATH_VARS)
		ldi zh,HIGH(MATH_VARS)
		clr a
		ldi b,MATH_VAR_BUF_SIZE
math_var_clear_loop:
		st z+,a
		dec b
		brne math_var_clear_loop

		pop zh
		pop zl
		pop b
		pop a
		ret
.endif
#endif

; ### Loads Accu (A=R[X])
; >a: The Index of the Register X
; <A: (loaded)
#ifndef MATH_FN_VAR_LOAD_A
#define MATH_FN_VAR_LOAD_A
.if MATH_VAR_NUM>0
math_var_load_a:
		push a
		push zl
		push zh
		
		cpi a,MATH_VAR_NUM
		brlo math_var_load_a_do
math_var_load_a_failed:
		clr a0
		clr a1
		clr a2
		clr a3
		rjmp math_var_load_a_exit
math_var_load_a_do:
		ldi zl,LOW(MATH_VARS)
		ldi zh,HIGH(MATH_VARS)
		lsl a
		lsl a
		add zl,a
		adc zh,zero
		ld a0,z+
		ld a1,z+
		ld a2,z+
		ld a3,z

math_var_load_a_exit:
		pop zh
		pop zl
		pop a
		ret
.endif
#endif

; ### Saves Accu (R[X]=A)
; >a: The Index of the Register X
; <A: (saved)
#ifndef MATH_FN_VAR_SAVE_A
#define MATH_FN_VAR_SAVE_A
.if MATH_VAR_NUM>0
math_var_save_a:
		push a
		push zl
		push zh

		cpi a,MATH_VAR_NUM
		brsh math_var_save_a_exit
		ldi zl,LOW(MATH_VARS)
		ldi zh,HIGH(MATH_VARS)
		lsl a
		lsl a
		add zl,a
		adc zh,zero
		st z+,a0
		st z+,a1
		st z+,a2
		st z,a3

math_var_save_a_exit:
		pop zh
		pop zl
		pop a
		ret
.endif
#endif

; ### Loads Operand (O=R[X])
; >a: The Index of the Register X
; <O: (loaded)
#ifndef MATH_FN_VAR_LOAD_O
#define MATH_FN_VAR_LOAD_O
.if MATH_VAR_NUM>0
math_var_load_o:
		push a
		push zl
		push zh

		cpi a,MATH_VAR_NUM
		brlo math_var_load_o_do
math_var_load_o_failed:
		clr o0
		clr o1
		clr o2
		clr o3
		rjmp math_var_load_o_exit
math_var_load_o_do:
		ldi zl,LOW(MATH_VARS)
		ldi zh,HIGH(MATH_VARS)
		lsl a
		lsl a
		add zl,a
		adc zh,zero
		ld o0,z+
		ld o1,z+
		ld o2,z+
		ld o3,z

math_var_load_o_exit:
		pop zh
		pop zl
		pop a
		ret
.endif
#endif

; ### Saves Operand (R[X]=O)
; >a: The Index of the Register X
; <O: (saved)
#ifndef MATH_FN_VAR_SAVE_O
#define MATH_FN_VAR_SAVE_O
.if MATH_VAR_NUM>0
math_var_save_o:
		push a
		push zl
		push zh

		cpi a,MATH_VAR_NUM
		brsh math_var_save_o_exit
		ldi zl,LOW(MATH_VARS)
		ldi zh,HIGH(MATH_VARS)
		lsl a
		lsl a
		add zl,a
		adc zh,zero
		st z+,o0
		st z+,o1
		st z+,o2
		st z,o3

math_var_save_o_exit:
		pop zh
		pop zl
		pop a
		ret
.endif
#endif

; ### Clear Accu (A=0)
; <A: (cleared)
#ifndef MATH_FN_CLEAR_A
#define MATH_FN_CLEAR_A
math_clear_a:
		clr a0
		clr a1
		clr a2
		clr a3
		ret
#endif

; ### Load Accu (A=K)
; >a: The Byte #0 (LSB)
; >b: The Byte #1
; >c: The Byte #2
; >d: The Byte #3 (MSB)
; <A: (loaded)
#ifndef MATH_FN_LOAD_A
#define MATH_FN_LOAD_A
math_load_a:
		mov a0,a
		mov a1,b
		mov a2,c
		mov a3,d
		ret
#endif

; ### Clear Operand (O=0)
; <O: (cleared)
#ifndef MATH_FN_CLEAR_O
#define MATH_FN_CLEAR_O
math_clear_o:
		clr o0
		clr o1
		clr o2
		clr o3
		ret
#endif

; ### Load Operand (O=K)
; >a: The Byte #0 (LSB)
; >b: The Byte #1
; >c: The Byte #2
; >d: The Byte #3 (MSB)
; <O: (loaded)
#ifndef MATH_FN_LOAD_O
#define MATH_FN_LOAD_O
math_load_o:
		mov o0,a
		mov o1,b
		mov o2,c
		mov o3,d
		ret
#endif

; ### Copy (O=A)
; <O: (copied)
#ifndef MATH_FN_COPY
#define MATH_FN_COPY
math_copy:
		mov o0,a0
		mov o1,a1
		mov o2,a2
		mov o3,a3
		ret
#endif

; ### Swap (O<=>A)
; <A&O: (swapped)
#ifndef MATH_FN_SWAP
#define MATH_FN_SWAP
math_swap:
		mov t0,o0
		mov t1,o1
		mov t2,o2
		mov t3,o3
		mov o0,a0
		mov o1,a1
		mov o2,a2
		mov o3,a3
		mov a0,t0
		mov a1,t1
		mov a2,t2
		mov a3,t3
		ret
#endif

; ### Compare (A&O)
; <Carry: The Carry-Flag.
#ifndef MATH_FN_COMPARE
#define MATH_FN_COMPARE
math_cmp:
		cp a0,o0
		cpc a1,o1
		cpc a2,o2
		cpc a3,o3
		ret
#endif

; ### Negate Accu (A=-A)
; <>A: The Result 
#ifndef MATH_FN_NEGATE_A
#define MATH_FN_NEGATE_A
math_neg_a:
		push a

		com a0
        com a1
        com a2
        com a3
		ldi a,-1
		sub a0,a
		sbc a1,a
		sbc a2,a
		sbc a3,a

		pop a
		ret
#endif

; ### Negate Accu (O=-O)
; <>O: The Result 
#ifndef MATH_FN_NEGATE_O
#define MATH_FN_NEGATE_O
math_neg_o:
		push a

		com o0
        com o1
        com o2
        com o3
		ldi a,-1
		sub o0,a
		sbc o1,a
		sbc o2,a
		sbc o3,a

		pop a
		ret
#endif

; ### Add Operand (A=A+O)
; >O: The Operand
; <>A: The Result 
; <carry: Overflow-Bit
#ifndef MATH_FN_ADD
#define MATH_FN_ADD
math_add:
		add a0,o0
		adc a1,o1
		adc a2,o2
		adc a3,o3
		ret
#endif

; ### Subtract Operand (A=A-O)
; >O: The Operand
; <>A: The result 
; <carry: Overflow-Bit
#ifndef MATH_FN_SUB
#define MATH_FN_SUB
math_sub:
		sub a0,o0
		sbc a1,o1
		sbc a2,o2
		sbc a3,o3
		ret
#endif

; ### (Signed) Multiplicate Operand (A=A*O)
; !!! only 32Bit * 16Bit !!!
; >O: The Operand
; <>A: The Result 
#ifndef MATH_FN_MUL
#define MATH_FN_MUL
math_mul:
		push a
		push b

		clr a								; Check Accu for Sign -> Negate
		sbrs a3,7							
		rjmp math_mul_chk1
		ser a
		rcall math_neg_a 
math_mul_chk1:
		clr b								; Check Operand for Sign -> Negate
		sbrs o3,7
		rjmp math_mul_chk2
		ser b
		rcall math_neg_o
math_mul_chk2:
		rcall math_umul						; Perform Unsigned Multiplication
		cpse a,b							
		rcall math_neg_a					; Set Resulting Sign		

		pop b
		pop a
		ret
#endif

; ### (Unsigned) Multiplicate Operand (A=A*O)
; !!! only 32Bit * 16Bit !!!
; >O: The Operand
; <>A: The Result 
#ifndef MATH_FN_UMUL
#define MATH_FN_UMUL
math_umul:
		push a

		cp a3,zero
		cpc a3,a2
		breq math_umul_start				; One Operand must be below 65536
		mov	t0,a0							; Swap A <-> B
		mov	a0,o0
		mov	o0,t0
		mov	t0,a1
		mov	a1,o1
		mov	o1,t0
		mov	o2,a2
		mov	o3,a3
		clr	a2
		clr	a3								; A(16Bit) & B(32Bit)
							
math_umul_start:							; a3,2,1,0 = a1,0 * b3,2,1,0
		ldi	a,16
		clr	t0
		clr	t1
		ror	a1
		ror	a0
math_umul_loop:	
		brcc math_umul_shift
		add	a2,o0
		adc	a3,o1
		adc	t0,o2
		adc	t1,o3
math_umul_shift:	
		ror	t1
		ror	t0
		ror	a3
		ror	a2
		ror	a1
		ror	a0
		dec	a
		brne math_umul_loop

		pop a
		ret				
#endif

; ### (Signed) Divisioin Operand (A=A/O)
; >O: The Operand
; <>A: The Result 
#ifndef MATH_FN_DIV
#define MATH_FN_DIV
math_div:
		push a
		push b

		clr a								; Check Accu for Sign -> Negate
		sbrs a3,7							
		rjmp math_div_chk1
		ser a
		rcall math_neg_a 
math_div_chk1:
		clr b								; Check Operand for Sign -> Negate
		sbrs o3,7
		rjmp math_div_chk2
		ser b
		rcall math_neg_o
math_div_chk2:
		rcall math_udiv						; Perform Unsigned Division
		cpse a,b							
		rcall math_neg_a					; Set Resulting Sign		

		pop b
		pop a
		ret
#endif

; ### (Unsigned) Divisioin Operand (A=A/O)
; >O: The Operand
; <>A: The Result 
#ifndef MATH_FN_UDIV
#define MATH_FN_UDIV
math_udiv:
		push a

		clr	t1
		tst	o3
		breq math_udiv_udi10
		ldi a,8
		mov	t0,a
math_udiv_udi1:	
		lsl	a0
		rol	a1
		rol	a2
		rol	a3
		rol	t1
		cp a1,o0
		cpc	a2,o1
		cpc	a3,o2
		cpc	t1,o3
		brcs math_udiv_udi2
		sub	a1,o0
		sbc	a2,o1
		sbc	a3,o2
		sbc	t1,o3
		inc	a0
math_udiv_udi2:	
		dec	t0
		brne math_udiv_udi1
		mov	o0,a1
		clr	a1
		mov	o1,a2
		clr	a2
		mov	o2,a3
		clr	a3
		mov	o3,t1
		rjmp math_udiv_exit

math_udiv_udi10:	
		tst	o2
		breq math_udiv_udi20
		ldi	a,16
		mov t0,a
math_udiv_udi11:	
		lsl	a0
		rol	a1
		rol	a2
		rol	a3
		rol	t1
		brcs math_udiv_udi12
		cp a2,o0
		cpc	a3,o1
		cpc	t1,o2
		brcs math_udiv_udi13
math_udiv_udi12:	
		sub	a2,o0
		sbc	a3,o1
		sbc	t1,o2
		inc	a0
math_udiv_udi13:	
		dec	t0
		brne math_udiv_udi11
		mov	o0,a2
		clr	a2
		mov	o1,a3
		clr	a3
		mov	o2,t1
		rjmp math_udiv_exit
 
math_udiv_udi20:	
		tst	o1
		breq math_udiv_udi30
		ldi	a,24
		mov t0,a
math_udiv_udi21:	
		lsl	a0
		rol	a1
		rol	a2
		rol	a3
		rol	t1
		brcs math_udiv_udi22
		cp a3,o0
		cpc	t1,o1
		brcs math_udiv_udi23
math_udiv_udi22:	
		sub	a3,o0
		sbc	t1,o1
		inc	a0
math_udiv_udi23:	
		dec	t0
		brne math_udiv_udi21
		mov	o0,a3
		clr	a3
		mov	o1,t1
		rjmp math_udiv_exit
 
math_udiv_udi30:
		ldi a,32	
		mov t0,a
math_udiv_udi31:
		lsl	a0
		rol	a1
		rol	a2
		rol	a3
		rol	t1
		brcs math_udiv_udi32
		cp t1,o0
		brcs math_udiv_udi33
math_udiv_udi32:
		sub	t1,o0
		inc	a0
math_udiv_udi33:	
		dec	t0
		brne math_udiv_udi31
		mov	o0,t1
	
math_udiv_exit:		
		pop a
		ret
#endif

; ### Shift Left
; >a: The Number of Bits to shift
; <>A: The Result 
#ifndef MATH_FN_SHL
#define MATH_FN_SHL
math_shl:
		push a

		tst a
		breq math_shl_exit
math_shl_loop:
		lsl a0
		rol a1
		rol a2
		rol a3
		dec a
		brne math_shl_loop

math_shl_exit:
		pop a
		ret
#endif

; ### Shift Right
; >a: The Number of Bits to shift
; <>A: The Result 
#ifndef MATH_FN_SHR
#define MATH_FN_SHR
math_shr:
		push a

		tst a
		breq math_shr_exit
math_shr_loop:
		lsr a3
		ror a2
		ror a1
		ror a0
		dec a
		brne math_shr_loop

math_shr_exit:
		pop a
		ret
#endif

; *****************************************************************************
; *** Private functions                                                     ***
; *****************************************************************************

#endif ; __MATH_INC
